#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include "config.h"

#ifndef HAVE_GETOPT
#include "getopt.c"
#endif

#define BLOCKSIZE 1024

#define HELP "elffilter [-i input-file] [-o output-file] [-r] [-d] search ...\n"

struct indext {
    off_t offset;
    long int length;
    time_t time;
};

char *getfield(char *line, int field);

int main(int argc, char *argv[])
{
    int c;
    long int size = 0, slide = 0, len = 0;
    char *line = NULL;
    char *outfile = NULL, *infile = NULL;
    char *field;
    int opt;
    FILE *fileout;
    FILE *filein;
    struct tm fromdate;
    time_t fromtime;
    struct tm todate;
    time_t totime;
    struct tm timedate;
    time_t timetime;
    char *str1;
    char *str2;

#define SCANBYRESULT 1
#define SCANBYDATE 2
#define SCANBYURI 3

    int scan = SCANBYURI;

    while ((opt = getopt(argc, argv, "i:o:rdhv")) != EOF) {
	switch (opt) {
	case 'i': {
	    infile = optarg;
	    break;
	}
	case 'o':
	    outfile = optarg;
	    break;

	case 'r':
	    scan = SCANBYRESULT;
	    break;

	case 'd':
	    scan = SCANBYDATE;
	    break;

	case 'h':
	    printf( HELP );
	    exit(0);

	case 'v':
	    printf("elffilter - " PACKAGE " " VERSION "\n");
	    exit(0);
	}
    }

    /* open up file for output */
    if (outfile) {
	if (!(fileout = fopen(outfile, "w"))) {
	    perror(outfile);
	    exit(1);
	}
    }
    else
	fileout = stdout;

    /* open up and read input file */
    if (infile) {
	if (!(filein = fopen(infile, "r"))) {
	    perror(infile);
	    exit(1);
	}
    }
    else
	filein = stdin;

    /* were we given any arguments to search for? */
    if (optind == argc) {
	printf( HELP );
	exit(0);
    }


    /* we loop through each line, passing the line to the outfile
     * if one of the search parameters is matched, it is included
     * in the output, otherwise the line is ignored.
     */

    switch (scan) {
    case SCANBYDATE: {

	/* default dates and times */
	fromdate.tm_year = 0;
	fromdate.tm_mon = 0;
	fromdate.tm_mday = 1;
	fromdate.tm_hour = 0;
	fromdate.tm_min = 0;
	fromdate.tm_sec = 0;
	todate.tm_year = 0;
	todate.tm_mon = 11;
	todate.tm_mday = 0;
	todate.tm_hour = 23;
	todate.tm_min = 59;
	todate.tm_sec = 59;

	/* read from date from the command line */
	opt = optind;
	switch (strlen(argv[opt])) {
	case 14 : {
	    fromdate.tm_sec = ((argv[opt])[12]-'0')*10+((argv[opt])[13]-'0');
	}
	case 12 : {
	    fromdate.tm_min = ((argv[opt])[10]-'0')*10+((argv[opt])[11]-'0');
	}
	case 10 : {
	    fromdate.tm_hour = ((argv[opt])[8]-'0')*10+((argv[opt])[9]-'0');
	}
	case 8: {
	    fromdate.tm_mday = ((argv[opt])[6]-'0')*10+((argv[opt])[7]-'0');
	}
	case 6 : {
	    fromdate.tm_mon = ((argv[opt])[4]-'0')*10+((argv[opt])[5]-'0')-1;
	}
	case 4 : {
	    fromdate.tm_year = ((argv[opt])[0]-'0')*1000+((argv[opt])[1]-'0')*100+((argv[opt])[2]-'0')*10+((argv[opt])[3]-'0')-1900;
	    todate.tm_year = fromdate.tm_year;
	    break;
	}
	default : {
	    printf("elffilter: Date format error. Should be YYYY[MM[DD[HH[MM[SS]]]]] [YYYY[MM[DD[HH[MM[SS]]]]]]. Seriously.\n");
	    exit(1);
	}
	}
	fromdate.tm_isdst = 0;

	/* read to date from the command line */
	opt++;
	if (argc > opt) {
	    switch (strlen(argv[opt])) {
	    case 14 : {
		todate.tm_sec = ((argv[opt])[12]-'0')*10+((argv[opt])[13]-'0');
	    }
	    case 12 : {
		todate.tm_min = ((argv[opt])[10]-'0')*10+((argv[opt])[11]-'0');
	    }
	    case 10 : {
		todate.tm_hour = ((argv[opt])[8]-'0')*10+((argv[opt])[9]-'0');
	    }
	    case 8: {
		todate.tm_mday = ((argv[opt])[6]-'0')*10+((argv[opt])[7]-'0');
	    }
	    case 6 : {
		todate.tm_mon = ((argv[opt])[4]-'0')*10+((argv[opt])[5]-'0')-1;
	    }
	    case 4 : {
		todate.tm_year = ((argv[opt])[0]-'0')*1000+((argv[opt])[1]-'0')*100+((argv[opt])[2]-'0')*10+((argv[opt])[3]-'0')-1900;
	    }
	    case 0 : {
		break;
	    }
	    default : {
		printf("Date format error. Should be YYYY[MM[DD[HH[MM[SS]]]]] [YYYY[MM[DD[HH[MM[SS]]]]]]. Seriously.\n");
		exit(1);
	    }
	    }

	}
	todate.tm_isdst = 0;
	/* fix to month if necessary */
	if (todate.tm_mday == 0)
	    todate.tm_mon++;

	/* convert dates */
	fromtime = mktime(&fromdate);
	totime = mktime(&todate);
/*printf("From: %d/%d/%d %d:%d:%d  To: %d/%d/%d %d:%d:%d\n", fromdate.tm_year, fromdate.tm_mon+1, fromdate.tm_mday, fromdate.tm_hour, fromdate.tm_min, fromdate.tm_sec, todate.tm_year, todate.tm_mon+1, todate.tm_mday, todate.tm_hour, todate.tm_min, todate.tm_sec);*/

	/* get each line of text until EOF */
	while ((c = fgetc(filein)) != EOF) {
	    /* enough space in buffer? no, make some more */
	    if (slide >= size) {
		size += BLOCKSIZE;
		if (!(line = (char *) realloc(line, size))) {
		    fprintf(stderr, "Memory allocation error\n");
		    exit(1);
		}
	    }

	    /* save character to buffer */
	    line[slide++] = (char) c;

	    /* reached an end of line? if so, deal with the line */
	    if (c == 10) {

		/* the line may be terminated with a CR, if so chop it off
		 * otherwise replace the LF with a NULL to indicate end of string */
		if ((slide > 0) && (line[slide - 1] == 13))
		    line[slide - 2] = 0;
		else
		    line[slide - 1] = 0;

		/* get the unix date from the line */
		field = getfield(line,4);
		timedate.tm_sec = (field[19]-'0')*10+(field[20]-'0');
		timedate.tm_min = (field[16]-'0')*10+(field[17]-'0');
		timedate.tm_hour = (field[13]-'0')*10+(field[14]-'0');
		timedate.tm_mday = (field[1]-'0')*10+(field[2]-'0');
		switch (field[4]) {
		case 'J' : {if(field[5] == 'a') timedate.tm_mon = 0; else if(field[6] == 'l') timedate.tm_mon = 6; else timedate.tm_mon = 5; break;}
		case 'F' : {timedate.tm_mon = 1; break;}
		case 'M' : {if(field[6] == 'y') timedate.tm_mon = 4; else timedate.tm_mon = 2; break;}
		case 'A' : {if(field[5] == 'p') timedate.tm_mon = 3; else timedate.tm_mon = 7; break;}
		case 'S' : {timedate.tm_mon = 8; break;}
		case 'O' : {timedate.tm_mon = 9; break;}
		case 'N' : {timedate.tm_mon = 10; break;}
		case 'D' : {timedate.tm_mon = 11; break;}
		}
		timedate.tm_year = (field[8]-'0')*1000+(field[9]-'0')*100+(field[10]-'0')*10+(field[11]-'0')-1900;
		timedate.tm_isdst = -1;
		free(field);

		field = getfield(line,5);
		timedate.tm_hour -= atoi(field);
		free(field);

		timetime = mktime(&timedate);
		if ((timetime > fromtime)&&(timetime < totime))
		    printf("%s\n", line);


		/* clean up */
		free(line);
		line = NULL;
		slide = 0;
		size = 0;
	    }
	}
	if (line)
	    free(line);

	break;
    }
    case SCANBYRESULT: {

	/* get each line of text until EOF */
	while ((c = fgetc(filein)) != EOF) {
	    /* enough space in buffer? no, make some more */
	    if (slide >= size) {
		size += BLOCKSIZE;
		if (!(line = (char *) realloc(line, size))) {
		    fprintf(stderr, "Memory allocation error\n");
		    exit(1);
		}
	    }

	    /* save character to buffer */
	    line[slide++] = (char) c;

	    /* reached an end of line? if so, deal with the line */
	    if (c == 10) {

		/* the line may be terminated with a CR, if so chop it off
		 * otherwise replace the LF with a NULL to indicate end of string */
		if ((slide > 0) && (line[slide - 1] == 13))
		    line[slide - 2] = 0;
		else
		    line[slide - 1] = 0;

		/* chop out the correct piece of the line */

		/* loop through each parameter, check the line against the parameter */
		opt = optind;
		while (opt < argc) {
		    if ((str1 = getfield(line, 7))) {
			if(!(strncmp(str1,argv[opt],strlen(argv[opt])))) {
			    printf("%s\n", line);
			    opt = argc;
			}
			free(str1);
		    }
		    opt++;
		}

		/* clean up */
		free(line);
		line = NULL;
		slide = 0;
		size = 0;
	    }
	}
	if (line)
	    free(line);

	break;
    }
    case SCANBYURI: {

	/* get each line of text until EOF */
	while ((c = fgetc(filein)) != EOF) {
	    /* enough space in buffer? no, make some more */
	    if (slide >= size) {
		size += BLOCKSIZE;
		if (!(line = (char *) realloc(line, size))) {
		    fprintf(stderr, "Memory allocation error\n");
		    exit(1);
		}
	    }

	    /* save character to buffer */
	    line[slide++] = (char) c;

	    /* reached an end of line? if so, deal with the line */
	    if (c == 10) {

		/* the line may be terminated with a CR, if so chop it off
		 * otherwise replace the LF with a NULL to indicate end of string */
		if ((slide > 0) && (line[slide - 1] == 13))
		    line[slide - 2] = 0;
		else
		    line[slide - 1] = 0;

		/* chop out the correct piece of the line */

		/* loop through each parameter, check the line against the parameter */
		opt = optind;
		while (opt < argc) {
		    len = strlen(argv[opt]);
		    if ((str1 = getfield(line, 6))) {
		        if ((str2 = getfield(str1+1, 2))) {
			    if( (!len) || (!(strncmp(str2,argv[opt],len)) && (argv[opt][len-1]=='/' || str2[len]=='/' || str2[len]==0)) ) {
				printf("%s\n", line);
				opt = argc;
			    }
			    free(str2);
			}
			free(str1);
		    }
		    opt++;
		}

		/* clean up */
		free(line);
		line = NULL;
		slide = 0;
		size = 0;
	    }
	}
	if (line)
	    free(line);
    
	break;
    }
    }


    /* close input file (but only if the file was a real file) */
    if (filein != stdin)
	fclose(filein);

    /* close out output file */
    if (fileout != stdout)
	fclose(fileout);

    exit(0);
}



/* get field number in a line */
/* buffer returned is size + 127 bytes so that it may be overwritten by the calling routine */
char *getfield(char *line, int field)
{
    char *buf;
    int start = 0, end = 0, slide = 0;
    int ws = 0;
    int br = 0;

    /* if ws is 0, we are finding text. if ws is not 0, we are finding whitespace */
    /* if br is not 0, we are inside a set of brackets */
    while (field && line[slide] && (line[slide] != 13)) {

	/* deal with brackets */
	if (line[slide] == '"') {
	    if (br)
		br = 0;
	    else
		br = 1;
	}

	/* find whitespace */
	switch (line[slide]) {
	case 8:
	case ' ':{
		if (!br && ws) {
		    ws = 0;
		}
		break;
	    }
	default:{
		if (!ws) {
		    field--;
		    ws = 1;
		    start = slide;
		}
	    }
	}
	slide++;
    }

    /* we found the start, now find the end */
    while (!end && line[slide] && (line[slide] != 13)) {

	/* find whitespace */
	switch (line[slide]) {
	case '"':{
		if (br)
		    br = 0;
		else
		    br = 1;
		break;
	    }
	case 8:
	case ' ':{
		if (!br) {
		    end = slide;
		}
		break;
	    }
	}
	slide++;
    }

    /* if !end, we found the last field */
    if (!end)
	end = slide;

    /* if field not found, return "" */
    if (field)
	start = end = 0;

    /* create a buffer for the field and return it - buffer contains an extra 127 bytes */
    if (!(buf = (char *) malloc(end - start + 128)))
	return (NULL);
    strncpy(buf, line + start, end - start);
    buf[end - start] = 0;
    return (buf);
}

